@@ -1,15 +1,17 @@
 #include "pgroonga.h"
 
+#include "pgrn_compatible.h"
 #include "pgrn_convert.h"
 #include "pgrn_global.h"
 #include "pgrn_groonga.h"
 #include "pgrn_jsonb.h"
 #include "pgrn_options.h"
 #include "pgrn_value.h"
+#include "pgrn_wal.h"
 
 #include <catalog/pg_type.h>
 #include <utils/builtins.h>
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 #	include <utils/jsonb.h>
 #endif
 
@@ -17,11 +19,11 @@
 
 #include <xxhash.h>
 
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 PG_FUNCTION_INFO_V1(pgroonga_match_jsonb);
 #endif
 
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 typedef struct
 {
 	grn_obj *pathsTable;
@@ -31,6 +33,7 @@ typedef struct
 
 typedef struct PGrnJSONBInsertData
 {
+	Relation index;
 	grn_obj *pathsTable;
 	grn_obj *valuesTable;
 	grn_obj *pathColumn;
@@ -43,8 +46,10 @@ typedef struct PGrnJSONBInsertData
 	grn_obj *valueIDs;
 	grn_obj key;
 	grn_obj components;
+	grn_obj valueKey;
 	grn_obj path;
 	grn_obj pathIDs;
+	grn_obj paths;
 	grn_obj value;
 	grn_obj type;
 } PGrnJSONBInsertData;
@@ -165,62 +170,79 @@ PGrnJSONBLookupSizeLexicon(Relation index,
 }
 
 static grn_obj *
-PGrnJSONBCreatePathsTable(const char *name)
+PGrnJSONBCreatePathsTable(Relation index, const char *name)
 {
-	return PGrnCreateTable(name,
+	return PGrnCreateTable(index,
+						   name,
 						   GRN_OBJ_TABLE_PAT_KEY,
-						   grn_ctx_at(ctx, GRN_DB_SHORT_TEXT));
+						   grn_ctx_at(ctx, GRN_DB_SHORT_TEXT),
+						   NULL,
+						   NULL);
 }
 
 static grn_obj *
-PGrnJSONBCreateTypesTable(const char *name)
+PGrnJSONBCreateTypesTable(Relation index, const char *name)
 {
-	return PGrnCreateTable(name,
+	return PGrnCreateTable(index,
+						   name,
 						   GRN_OBJ_TABLE_PAT_KEY,
-						   grn_ctx_at(ctx, GRN_DB_SHORT_TEXT));
+						   grn_ctx_at(ctx, GRN_DB_SHORT_TEXT),
+						   NULL,
+						   NULL);
 }
 
 static grn_obj *
-PGrnJSONBCreateValuesTable(const char *name)
+PGrnJSONBCreateValuesTable(Relation index, const char *name)
 {
-	return PGrnCreateTable(name,
+	return PGrnCreateTable(index,
+						   name,
 						   GRN_OBJ_TABLE_HASH_KEY,
-						   grn_ctx_at(ctx, GRN_DB_UINT64));
+						   grn_ctx_at(ctx, GRN_DB_UINT64),
+						   NULL,
+						   NULL);
 }
 
 static void
-PGrnJSONBCreateDataColumns(PGrnJSONBCreateData *jsonbData)
+PGrnJSONBCreateDataColumns(Relation index,
+						   PGrnJSONBCreateData *jsonbData)
 {
-	PGrnCreateColumn(jsonbData->valuesTable,
+	PGrnCreateColumn(index,
+					 jsonbData->valuesTable,
 					 "path",
 					 GRN_OBJ_COLUMN_SCALAR,
 					 jsonbData->pathsTable);
-	PGrnCreateColumn(jsonbData->valuesTable,
+	PGrnCreateColumn(index,
+					 jsonbData->valuesTable,
 					 "paths",
 					 GRN_OBJ_COLUMN_VECTOR,
 					 jsonbData->pathsTable);
 	{
-		grn_obj_flags flags = 0;
+		grn_column_flags flags = 0;
 		if (PGrnIsLZ4Available)
 			flags |= GRN_OBJ_COMPRESS_LZ4;
-		PGrnCreateColumn(jsonbData->valuesTable,
+		PGrnCreateColumn(index,
+						 jsonbData->valuesTable,
 						 "string",
 						 flags,
 						 grn_ctx_at(ctx, GRN_DB_LONG_TEXT));
 	}
-	PGrnCreateColumn(jsonbData->valuesTable,
+	PGrnCreateColumn(index,
+					 jsonbData->valuesTable,
 					 "number",
 					 0,
 					 grn_ctx_at(ctx, GRN_DB_FLOAT));
-	PGrnCreateColumn(jsonbData->valuesTable,
+	PGrnCreateColumn(index,
+					 jsonbData->valuesTable,
 					 "boolean",
 					 0,
 					 grn_ctx_at(ctx, GRN_DB_BOOL));
-	PGrnCreateColumn(jsonbData->valuesTable,
+	PGrnCreateColumn(index,
+					 jsonbData->valuesTable,
 					 "size",
 					 0,
 					 grn_ctx_at(ctx, GRN_DB_UINT32));
-	PGrnCreateColumn(jsonbData->valuesTable,
+	PGrnCreateColumn(index,
+					 jsonbData->valuesTable,
 					 "type",
 					 0,
 					 jsonbData->typesTable);
@@ -325,9 +347,11 @@ PGrnJSONBInsertDataInit(PGrnJSONBInsertData *data)
 
 	GRN_TEXT_INIT(&(data->key), 0);
 	GRN_TEXT_INIT(&(data->components), GRN_OBJ_VECTOR);
+	GRN_UINT64_INIT(&(data->valueKey), 0);
 	GRN_TEXT_INIT(&(data->path), 0);
 	GRN_RECORD_INIT(&(data->pathIDs), GRN_OBJ_VECTOR,
 					grn_obj_id(ctx, data->pathsTable));
+	GRN_TEXT_INIT(&(data->paths), GRN_OBJ_VECTOR);
 	GRN_VOID_INIT(&(data->value));
 	GRN_TEXT_INIT(&(data->type), GRN_OBJ_DO_SHALLOW_COPY);
 }
@@ -337,8 +361,10 @@ PGrnJSONBInsertDataFin(PGrnJSONBInsertData *data)
 {
 	GRN_OBJ_FIN(ctx, &(data->type));
 	GRN_OBJ_FIN(ctx, &(data->value));
+	GRN_OBJ_FIN(ctx, &(data->paths));
 	GRN_OBJ_FIN(ctx, &(data->pathIDs));
 	GRN_OBJ_FIN(ctx, &(data->path));
+	GRN_OBJ_FIN(ctx, &(data->valueKey));
 	GRN_OBJ_FIN(ctx, &(data->components));
 	GRN_OBJ_FIN(ctx, &(data->key));
 }
@@ -477,6 +503,9 @@ PGrnJSONBInsertValueSet(PGrnJSONBInsertData *data,
 	uint64_t key;
 	grn_id valueID;
 	int added;
+	size_t nColumns = 3; /* _key, path?, paths, column?, type */
+	bool setPath;
+	PGrnWALData *walData = NULL;
 
 	key = PGrnJSONBInsertGenerateKey(data, column != NULL, typeName);
 	valueID = grn_table_add(ctx, data->valuesTable,
@@ -488,20 +517,74 @@ PGrnJSONBInsertValueSet(PGrnJSONBInsertData *data,
 
 	GRN_BULK_REWIND(&(data->path));
 	PGrnJSONGenerateCompletePath(&(data->components), &(data->path));
-	if (GRN_TEXT_LEN(&(data->path)) < GRN_TABLE_MAX_KEY_SIZE)
+	setPath = (GRN_TEXT_LEN(&(data->path)) < GRN_TABLE_MAX_KEY_SIZE);
+	if (setPath)
+		nColumns++;
+
+	if (column)
+		nColumns++;
+
+	if (data->index)
+		walData = PGrnWALStart(data->index);
+
+	PGrnWALInsertStart(walData, data->valuesTable, nColumns);
+	if (walData)
+	{
+		GRN_UINT64_SET(ctx, &(data->valueKey), key);
+		PGrnWALInsertKey(walData, &(data->valueKey));
+	}
+
+	if (setPath)
+	{
 		grn_obj_set_value(ctx, data->pathColumn, valueID,
 						  &(data->path), GRN_OBJ_SET);
+		PGrnWALInsertColumn(walData, data->pathColumn, &(data->path));
+	}
 
 	PGrnJSONBInsertGeneratePaths(data);
 	grn_obj_set_value(ctx, data->pathsColumn, valueID,
 					  &(data->pathIDs), GRN_OBJ_SET);
+	if (walData)
+	{
+		unsigned int i, n;
+
+		GRN_BULK_REWIND(&(data->paths));
+		n = GRN_BULK_VSIZE(&(data->pathIDs)) / sizeof(grn_id);
+		for (i = 0; i < n; i++)
+		{
+			grn_id pathID;
+			char path[GRN_TABLE_MAX_KEY_SIZE];
+			int pathSize;
+
+			pathID = GRN_RECORD_VALUE_AT(&(data->pathIDs), i);
+			pathSize = grn_table_get_key(ctx,
+										 data->pathsTable,
+										 pathID,
+										 path,
+										 GRN_TABLE_MAX_KEY_SIZE);
+			grn_vector_add_element(ctx,
+								   &(data->paths),
+								   path,
+								   pathSize,
+								   0,
+								   GRN_DB_SHORT_TEXT);
+		}
+		PGrnWALInsertColumn(walData, data->pathsColumn, &(data->paths));
+	}
 
 	if (column)
+	{
 		grn_obj_set_value(ctx, column, valueID, &(data->value), GRN_OBJ_SET);
+		PGrnWALInsertColumn(walData, column, &(data->value));
+	}
 
 	GRN_TEXT_SETS(ctx, &(data->type), typeName);
 	grn_obj_set_value(ctx, data->typeColumn, valueID,
 					  &(data->type), GRN_OBJ_SET);
+	PGrnWALInsertColumn(walData, data->typeColumn, &(data->type));
+
+	PGrnWALInsertFinish(walData);
+	PGrnWALFinish(walData);
 }
 
 static void PGrnJSONBInsertContainer(JsonbIterator **iter,
@@ -527,10 +610,13 @@ PGrnJSONBInsertValue(JsonbIterator **iter,
 		break;
 	case jbvNumeric:
 	{
-		Datum numericInString =
+		Datum numericInString;
+		const char *numericInCString;
+
+		numericInString =
 			DirectFunctionCall1(numeric_out,
 								NumericGetDatum(value->val.numeric));
-		const char *numericInCString = DatumGetCString(numericInString);
+		numericInCString = DatumGetCString(numericInString);
 		grn_obj_reinit(ctx, &(data->value), GRN_DB_TEXT,
 					   GRN_OBJ_DO_SHALLOW_COPY);
 		GRN_TEXT_SETS(ctx, &(data->value), numericInCString);
@@ -618,31 +704,31 @@ PGrnJSONBInsertContainer(JsonbIterator **iter, PGrnJSONBInsertData *data)
 void
 PGrnInitializeJSONB(void)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 	PGrnJSONBCreateData data;
 
-	tmpPathsTable = PGrnJSONBCreatePathsTable(NULL);
-	tmpTypesTable = PGrnJSONBCreateTypesTable(NULL);
-	tmpValuesTable = PGrnJSONBCreateValuesTable(NULL);
+	tmpPathsTable = PGrnJSONBCreatePathsTable(InvalidRelation, NULL);
+	tmpTypesTable = PGrnJSONBCreateTypesTable(InvalidRelation, NULL);
+	tmpValuesTable = PGrnJSONBCreateValuesTable(InvalidRelation, NULL);
 
 	data.pathsTable = tmpPathsTable;
 	data.typesTable = tmpTypesTable;
 	data.valuesTable = tmpValuesTable;
-	PGrnJSONBCreateDataColumns(&data);
+	PGrnJSONBCreateDataColumns(InvalidRelation, &data);
 #endif
 }
 
 void
 PGrnFinalizeJSONB(void)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 	grn_obj_remove(ctx, tmpValuesTable);
 	grn_obj_remove(ctx, tmpTypesTable);
 	grn_obj_remove(ctx, tmpPathsTable);
 #endif
 }
 
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 static void
 PGrnJSONBCreateTables(PGrnCreateData *data,
 					  PGrnJSONBCreateData *jsonbData)
@@ -652,7 +738,8 @@ PGrnJSONBCreateTables(PGrnCreateData *data,
 		snprintf(jsonPathsTableName, sizeof(jsonPathsTableName),
 				 PGrnJSONPathsTableNameFormat,
 				 data->relNode, data->i);
-		jsonbData->pathsTable = PGrnJSONBCreatePathsTable(jsonPathsTableName);
+		jsonbData->pathsTable =
+			PGrnJSONBCreatePathsTable(data->index, jsonPathsTableName);
 		GRN_PTR_PUT(ctx, data->supplementaryTables, jsonbData->pathsTable);
 	}
 
@@ -661,7 +748,8 @@ PGrnJSONBCreateTables(PGrnCreateData *data,
 		snprintf(jsonTypesTableName, sizeof(jsonTypesTableName),
 				 PGrnJSONTypesTableNameFormat,
 				 data->relNode, data->i);
-		jsonbData->typesTable = PGrnJSONBCreateTypesTable(jsonTypesTableName);
+		jsonbData->typesTable =
+			PGrnJSONBCreateTypesTable(data->index, jsonTypesTableName);
 		GRN_PTR_PUT(ctx, data->supplementaryTables, jsonbData->typesTable);
 	}
 
@@ -670,7 +758,8 @@ PGrnJSONBCreateTables(PGrnCreateData *data,
 		snprintf(jsonValuesTableName, sizeof(jsonValuesTableName),
 				 PGrnJSONValuesTableNameFormat,
 				 data->relNode, data->i);
-		jsonbData->valuesTable = PGrnJSONBCreateValuesTable(jsonValuesTableName);
+		jsonbData->valuesTable =
+			PGrnJSONBCreateValuesTable(data->index, jsonValuesTableName);
 		GRN_PTR_PUT(ctx, data->supplementaryTables, jsonbData->valuesTable);
 	}
 }
@@ -682,7 +771,11 @@ PGrnJSONBCreateFullTextSearchIndexColumn(PGrnCreateData *data,
 	const char *tokenizerName = PGRN_DEFAULT_TOKENIZER;
 	const char *normalizerName = PGRN_DEFAULT_NORMALIZER;
 	char lexiconName[GRN_TABLE_MAX_KEY_SIZE];
+	grn_table_flags flags = GRN_OBJ_TABLE_PAT_KEY;
+	grn_obj *type;
 	grn_obj *lexicon;
+	grn_obj *tokenizer;
+	grn_obj *normalizer = NULL;
 
 	PGrnApplyOptionValues(data->index, &tokenizerName, &normalizerName);
 
@@ -692,20 +785,20 @@ PGrnJSONBCreateFullTextSearchIndexColumn(PGrnCreateData *data,
 	snprintf(lexiconName, sizeof(lexiconName),
 			 PGrnJSONValueLexiconNameFormat,
 			 "FullTextSearch", data->relNode, data->i);
-	lexicon = PGrnCreateTable(lexiconName,
-							  GRN_OBJ_TABLE_PAT_KEY,
-							  grn_ctx_at(ctx, GRN_DB_SHORT_TEXT));
+	type = grn_ctx_at(ctx, GRN_DB_SHORT_TEXT);
+	tokenizer = PGrnLookup(tokenizerName, ERROR);
+	if (!PGrnIsNoneValue(normalizerName))
+		normalizer = PGrnLookup(normalizerName, ERROR);
+	lexicon = PGrnCreateTable(data->index,
+							  lexiconName,
+							  flags,
+							  type,
+							  tokenizer,
+							  normalizer);
 	GRN_PTR_PUT(ctx, data->lexicons, lexicon);
 
-	grn_obj_set_info(ctx, lexicon, GRN_INFO_DEFAULT_TOKENIZER,
-					 PGrnLookup(tokenizerName, ERROR));
-	if (!PGrnIsNoneValue(normalizerName))
-	{
-		grn_obj_set_info(ctx, lexicon, GRN_INFO_NORMALIZER,
-						 PGrnLookup(normalizerName, ERROR));
-	}
-
-	PGrnCreateColumn(lexicon,
+	PGrnCreateColumn(data->index,
+					 lexicon,
 					 PGrnIndexColumnName,
 					 GRN_OBJ_COLUMN_INDEX | GRN_OBJ_WITH_POSITION,
 					 jsonbData->valuesTable);
@@ -724,9 +817,15 @@ PGrnJSONBCreateIndexColumn(PGrnCreateData *data,
 	snprintf(lexiconName, sizeof(lexiconName),
 			 PGrnJSONValueLexiconNameFormat,
 			 typeName, data->relNode, data->i);
-	lexicon = PGrnCreateTable(lexiconName, tableType, type);
+	lexicon = PGrnCreateTable(data->index,
+							  lexiconName,
+							  tableType,
+							  type,
+							  NULL,
+							  NULL);
 	GRN_PTR_PUT(ctx, data->lexicons, lexicon);
-	PGrnCreateColumn(lexicon,
+	PGrnCreateColumn(data->index,
+					 lexicon,
 					 PGrnIndexColumnName,
 					 GRN_OBJ_COLUMN_INDEX,
 					 jsonbData->valuesTable);
@@ -736,11 +835,13 @@ static void
 PGrnJSONBCreateIndexColumns(PGrnCreateData *data,
 							PGrnJSONBCreateData *jsonbData)
 {
-	PGrnCreateColumn(jsonbData->valuesTable,
+	PGrnCreateColumn(data->index,
+					 jsonbData->valuesTable,
 					 PGrnIndexColumnName,
 					 GRN_OBJ_COLUMN_INDEX,
 					 data->sourcesTable);
-	PGrnCreateColumn(jsonbData->pathsTable,
+	PGrnCreateColumn(data->index,
+					 jsonbData->pathsTable,
 					 PGrnIndexColumnName,
 					 GRN_OBJ_COLUMN_INDEX | GRN_OBJ_WITH_SECTION,
 					 jsonbData->valuesTable);
@@ -774,7 +875,7 @@ PGrnJSONBCreateIndexColumns(PGrnCreateData *data,
 bool
 PGrnAttributeIsJSONB(Oid id)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 	return id == JSONBOID;
 #else
 	return false;
@@ -784,7 +885,7 @@ PGrnAttributeIsJSONB(Oid id)
 void
 PGrnJSONBCreate(PGrnCreateData *data)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 	PGrnJSONBCreateData jsonbData;
 
 	if (data->desc->natts != 1)
@@ -797,17 +898,15 @@ PGrnJSONBCreate(PGrnCreateData *data)
 	}
 
 	PGrnJSONBCreateTables(data, &jsonbData);
-	PGrnJSONBCreateDataColumns(&jsonbData);
+	PGrnJSONBCreateDataColumns(data->index, &jsonbData);
 	PGrnJSONBCreateIndexColumns(data, &jsonbData);
-	data->forFullTextSearch = false;
-	data->forRegexpSearch = false;
 	data->attributeTypeID = grn_obj_id(ctx, jsonbData.valuesTable);
 	data->attributeFlags = GRN_OBJ_VECTOR;
 	PGrnCreateDataColumn(data);
 #endif
 }
 
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 static void
 PGrnJSONBValueSetSource(Relation index,
 						grn_obj *jsonValuesTable,
@@ -838,7 +937,7 @@ PGrnJSONBValueSetSource(Relation index,
 	}
 
 	source = PGrnLookupColumn(jsonValuesTable, columnName, ERROR);
-	PGrnIndexColumnSetSource(indexColumn, source);
+	PGrnIndexColumnSetSource(index, indexColumn, source);
 
 	grn_obj_unlink(ctx, source);
 	grn_obj_unlink(ctx, indexColumn);
@@ -856,22 +955,22 @@ PGrnJSONBSetSources(Relation index,
 	{
 		grn_obj *source;
 		grn_obj *indexColumn;
+		grn_obj *sourceIDs;
 
-		GRN_BULK_REWIND(&(buffers->sourceIDs));
+		sourceIDs = &(buffers->sourceIDs);
+		GRN_BULK_REWIND(sourceIDs);
 
 		source = PGrnLookupColumn(jsonValuesTable, "path", ERROR);
-		GRN_RECORD_PUT(ctx, &(buffers->sourceIDs), grn_obj_id(ctx, source));
+		GRN_RECORD_PUT(ctx, sourceIDs, grn_obj_id(ctx, source));
 		grn_obj_unlink(ctx, source);
 
 		source = PGrnLookupColumn(jsonValuesTable, "paths", ERROR);
-		GRN_RECORD_PUT(ctx, &(buffers->sourceIDs), grn_obj_id(ctx, source));
+		GRN_RECORD_PUT(ctx, sourceIDs, grn_obj_id(ctx, source));
 		grn_obj_unlink(ctx, source);
 
 		indexColumn = PGrnLookupColumn(jsonPathsTable, PGrnIndexColumnName,
 									   ERROR);
-		grn_obj_set_info(ctx, indexColumn, GRN_INFO_SOURCE,
-						 &(buffers->sourceIDs));
-		grn_obj_unlink(ctx, indexColumn);
+		PGrnIndexColumnSetSourceIDs(index, indexColumn, sourceIDs);
 	}
 
 	PGrnJSONBValueSetSource(index, jsonValuesTable, "string", "String",
@@ -892,7 +991,7 @@ PGrnJSONBSetSources(Relation index,
 grn_obj *
 PGrnJSONBSetSource(Relation index, unsigned int i)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 	grn_obj *jsonValuesTable;
 	grn_obj *indexColumn;
 
@@ -908,7 +1007,7 @@ PGrnJSONBSetSource(Relation index, unsigned int i)
 #endif
 }
 
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 static void
 PGrnJSONBDeleteValues(grn_obj *valuesTable, grn_obj *valueIDs)
 {
@@ -939,6 +1038,7 @@ pgroonga_match_jsonb(PG_FUNCTION_ARGS)
 	grn_obj *result = NULL;
 	bool matched = false;
 
+	data.index = NULL;
 	data.pathsTable  = tmpPathsTable;
 	data.valuesTable = tmpValuesTable;
 	GRN_PTR_INIT(&valueIDs, GRN_OBJ_VECTOR, grn_obj_id(ctx, data.valuesTable));
@@ -951,19 +1051,21 @@ pgroonga_match_jsonb(PG_FUNCTION_ARGS)
 	PG_TRY();
 	{
 		GRN_EXPR_CREATE_FOR_QUERY(ctx, tmpValuesTable, filter, dummy_variable);
-		PGrnCheck("pgroonga: match_jsonb: failed to create expression object");
+		PGrnCheck("match_jsonb: failed to create expression object");
 		grn_expr_parse(ctx, filter,
 					   VARDATA_ANY(query),
 					   VARSIZE_ANY_EXHDR(query),
 					   NULL, GRN_OP_MATCH, GRN_OP_AND,
 					   GRN_EXPR_SYNTAX_SCRIPT);
-		PGrnCheck("pgroonga: match_jsonb: failed to parse query");
+		PGrnCheck("match_jsonb: failed to parse query: <%.*s>",
+				  (int)VARSIZE_ANY_EXHDR(query),
+				  VARDATA_ANY(query));
 		result = grn_table_create(ctx, NULL, 0, NULL,
 								  GRN_TABLE_HASH_KEY|GRN_OBJ_WITH_SUBREC,
 								  tmpValuesTable, NULL);
-		PGrnCheck("pgroonga: match_jsonb: failed to create result table");
+		PGrnCheck("match_jsonb: failed to create result table");
 		grn_table_select(ctx, tmpValuesTable, filter, result, GRN_OP_OR);
-		PGrnCheck("pgroonga: match_jsonb: failed to select");
+		PGrnCheck("match_jsonb: failed to select");
 	}
 	PG_CATCH();
 	{
@@ -987,34 +1089,110 @@ pgroonga_match_jsonb(PG_FUNCTION_ARGS)
 
 	PG_RETURN_BOOL(matched);
 }
+
+static void
+PGrnJSONBInsertJSONB(PGrnJSONBInsertData *data,
+					 Jsonb *jsonb)
+{
+	JsonbIterator *iter;
+
+	iter = JsonbIteratorInit(&(jsonb->root));
+	PGrnJSONBInsertContainer(&iter, data);
+}
+
+static void
+PGrnJSONBInsertRecord(Relation index,
+					  grn_obj *sourcesTable,
+					  grn_obj *sourcesCtidColumn,
+					  uint64_t packedCtid,
+					  PGrnJSONBInsertData *data)
+{
+	TupleDesc desc = RelationGetDescr(index);
+	grn_id id;
+	PGrnWALData *walData;
+	Form_pg_attribute attribute;
+	NameData *attributeName;
+	grn_obj *column;
+
+	id = grn_table_add(ctx, sourcesTable, NULL, 0, NULL);
+
+	walData = PGrnWALStart(index);
+	PGrnWALInsertStart(walData, NULL, desc->natts + 1);
+
+	GRN_UINT64_SET(ctx, &(buffers->ctid), packedCtid);
+	grn_obj_set_value(ctx, sourcesCtidColumn, id, &(buffers->ctid), GRN_OBJ_SET);
+	PGrnWALInsertColumn(walData, sourcesCtidColumn, &(buffers->ctid));
+
+	attribute = desc->attrs[0];
+	attributeName = &(attribute->attname);
+	column = PGrnLookupColumn(sourcesTable, attributeName->data, ERROR);
+	grn_obj_set_value(ctx, column, id, data->valueIDs, GRN_OBJ_SET);
+	PGrnCheck("failed to set column value: <%s>", attributeName->data);
+
+	if (walData)
+	{
+		grn_obj *valueKeys = &(buffers->jsonbValueKeys);
+		size_t i, nValueIDs;
+
+		GRN_BULK_REWIND(valueKeys);
+		nValueIDs = GRN_BULK_VSIZE(data->valueIDs) / sizeof(grn_id);
+		for (i = 0; i < nValueIDs; i++)
+		{
+			grn_id valueID;
+			uint64_t key;
+
+			valueID = GRN_RECORD_VALUE_AT(data->valueIDs, i);
+			grn_table_get_key(ctx,
+							  data->valuesTable,
+							  valueID,
+							  &key,
+							  sizeof(uint64_t));
+			GRN_UINT64_PUT(ctx, valueKeys, key);
+		}
+		PGrnWALInsertColumn(walData, column, valueKeys);
+	}
+
+	PGrnWALInsertFinish(walData);
+	PGrnWALFinish(walData);
+}
 #endif
 
 void
 PGrnJSONBInsert(Relation index,
+				grn_obj *sourcesTable,
+				grn_obj *sourcesCtidColumn,
 				Datum *values,
-				unsigned int nthValue,
-				grn_obj *valueIDs)
+				bool *isnull,
+				uint64_t packedCtid)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 	PGrnJSONBInsertData data;
+	unsigned int nthValue = 0;
 	Jsonb *jsonb;
-	JsonbIterator *iter;
 
+	data.index = index;
 	data.pathsTable  = PGrnJSONBLookupPathsTable(index, nthValue, ERROR);
 	data.valuesTable = PGrnJSONBLookupValuesTable(index, nthValue, ERROR);
-	data.valueIDs = valueIDs;
+	data.valueIDs = &(buffers->general);
 	grn_obj_reinit(ctx, data.valueIDs,
 				   grn_obj_id(ctx, data.valuesTable),
 				   GRN_OBJ_VECTOR);
 	PGrnJSONBInsertDataInit(&data);
+
 	jsonb = DatumGetJsonb(values[nthValue]);
-	iter = JsonbIteratorInit(&(jsonb->root));
-	PGrnJSONBInsertContainer(&iter, &data);
+	PGrnJSONBInsertJSONB(&data, jsonb);
+
+	PGrnJSONBInsertRecord(index,
+						  sourcesTable,
+						  sourcesCtidColumn,
+						  packedCtid,
+						  &data);
+
 	PGrnJSONBInsertDataFin(&data);
 #endif
 }
 
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 static void
 PGrnSearchBuildConditionJSONQuery(PGrnSearchData *data,
 								  grn_obj *subFilter,
@@ -1216,7 +1394,7 @@ PGrnJSONBBuildSearchCondition(PGrnSearchData *data,
 							  ScanKey key,
 							  grn_obj *targetColumn)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 	grn_obj *subFilter;
 
 	subFilter = PGrnLookup("sub_filter", ERROR);
@@ -1245,7 +1423,7 @@ PGrnJSONBBuildSearchCondition(PGrnSearchData *data,
 #endif
 }
 
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 static void
 PGrnJSONValuesUpdateDeletedID(grn_obj *jsonValuesTable,
 							  grn_obj *values,
@@ -1333,7 +1511,7 @@ PGrnJSONValuesDeleteBulk(grn_obj *jsonValuesTable,
 void
 PGrnJSONBBulkDeleteInit(PGrnJSONBBulkDeleteData *data)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 	Relation index = data->index;
 	TupleDesc desc;
 	Form_pg_attribute attribute;
@@ -1366,7 +1544,7 @@ PGrnJSONBBulkDeleteInit(PGrnJSONBBulkDeleteData *data)
 void
 PGrnJSONBBulkDeleteRecord(PGrnJSONBBulkDeleteData *data)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 	if (!data->isJSONBAttribute)
 		return;
 
@@ -1385,7 +1563,7 @@ PGrnJSONBBulkDeleteRecord(PGrnJSONBBulkDeleteData *data)
 void
 PGrnJSONBBulkDeleteFin(PGrnJSONBBulkDeleteData *data)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 	if (!data->isJSONBAttribute)
 		return;
 
@@ -1401,7 +1579,7 @@ PGrnJSONBBulkDeleteFin(PGrnJSONBBulkDeleteData *data)
 #endif
 }
 
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 static bool
 PGrnRemoveJSONValueLexicon(const char *typeName, unsigned int relationID)
 {
@@ -1416,7 +1594,7 @@ PGrnRemoveJSONValueLexicon(const char *typeName, unsigned int relationID)
 void
 PGrnJSONBRemoveUnusedTables(Oid relationFileNodeID)
 {
-#ifdef JSONBOID
+#ifdef PGRN_SUPPORT_JSONB
 		PGrnRemoveJSONValueLexicon("FullTextSearch", relationFileNodeID);
 		PGrnRemoveJSONValueLexicon("String", relationFileNodeID);
 		PGrnRemoveJSONValueLexicon("Number", relationFileNodeID);
